/*
This sketch illustrates ideas from the course "2D Animation and Interaction"
available at www.imaginary-institute.com, and my book "Processing for
Artists and Designers," published by AK Peters. For more of my Processing
programs, visit www.glassner.com or www.imaginary-institute.com
---
I draw a bunch of little colored dots falling from the top of the window downward. With
each step they get affected by the winds, which are circular vortexes that slowly bounce
around the window. The overall color palette changes slowly over time. Drawing more dots
(with the variable NumSparks) generates a denser and smoother pattern. Creating more wind
patterns (with the variable NumGusts) makes the dots swirl around more and in more complex 
ways as they fall. You can mess around with both of these numbers any way you like. Generally
speaking, more sparks or more gusts will slow things down but the pattern will look richer
in some way; fewer of these objects will draw faster and give a slightly more chaotic look.
---
Andrew Glassner, version 2, April 2014
*/
// User variables
int NumSparks = 1000;  // how many colored dots are falling at any given time
int NumGusts = 18;     // how many swirling winds are pushing the dots around
float HueChangeSpeed = .02; // how fast the colors change over time. Larger=faster.
// The two global variables that drive the program
Grain [] Sparks;  // the Grain objects that fall down the screen
Wind [] Gusts;    // the Wind objects that push the Grains around
void setup() {
   size(600, 400);
   Sparks = new Grain[NumSparks];
   for (int i=0; i<Sparks.length; i++) {
      Sparks[i] = new Grain();
   }
   Gusts = new Wind[NumGusts];
   for (int i=0; i<Gusts.length; i++) {
      Gusts[i] = new Wind();
   }
}
// Draw each Spark and update it, then update all the Gusts.
void draw() {
   for (int i=0; i<Sparks.length; i++) {
      Sparks[i].render();
      Sparks[i].move();
   }
   for (int i=0; i<Gusts.length; i++) {
      Gusts[i].update();
   }
}
//----------------------------- GRAIN CLASS
class Grain {
   PVector center;       // center location
   float radius;         // size in pixels
   float ch, cs, cb;     // HSB values of the dot
   PVector oldVelocity;  // movement in last frame
   PVector oldCenter;    // position in last frame
   Grain() {
      fillGrain();   // constructor fills the grain with new values
      /* fillGrain() starts each Grain vertically at the top of the window.
      But when we're first creating all the Grains, we don't want them all
      starting at the same height, or they'll drop down in unison like
      a curtain, which doesn't look very nice to me. So I tweak the very
      start of the program by moving each Grain to a random position 
      somewhere above the window, delaying its initial arrival.
      */
      center.y = random(-height, 0);
   }
   // Fill in the Grain with random values.  
   void fillGrain() {
      center = new PVector(random(0, width), 0);                      // start somewhere at the top
      oldCenter = new PVector(center.x, center.y);                    // the "old value" is just this one
      oldVelocity = new PVector(random(-.02,.02), random(.4, 1.8));   // the "old velcity" is mostly downward
      radius = random(0.25, 2.0);                                     // pick a small random size for the dot
      float colorSets [] = {   // five hand-picked HSB colors I like
         35, 238, 133,
         42, 231, 210,
         21, 232, 235,
         18, 204, 115,
          9, 248, 210, };
      int colorSteps = int(colorSets.length/3)-1;    // how many sets of colors do we have?
      float setWidth = 1.0*width/colorSteps;         // space them out equally horizontally
      int setNumber = int(center.x/setWidth);        // find out which color corresponds to this position in x
      float a = (center.x - (setNumber * setWidth))/setWidth;               // blend factor between color sets
      ch = lerp(colorSets[(setNumber*3)], colorSets[(setNumber*3)+3], a);   // blend the hues to get a new hue
      cs = lerp(colorSets[(setNumber*3)+1], colorSets[(setNumber*3)+4], a); // blend saturations
      cb = lerp(colorSets[(setNumber*3)+2], colorSets[(setNumber*3)+5], a); // blend brightnesses
      ch = (ch + (frameCount * HueChangeSpeed)) % 255.0;  // change the overall palette over time
      cs = constrain(cs+random(-10, 10), 0, 255);  // wiggle the saturation, but make sure it stays (0-255)
      cb = constrain(cb+random(-10, 10), 0, 255);  // wiggle the brightness, but make sure it stays (0-255)
   }
   // draw this Grain as a short stroke from the oldCenter to center
   void render() {
      strokeWeight(radius);      
      colorMode(HSB);
      stroke(ch, cs, cb);
      line(oldCenter.x, oldCenter.y, center.x, center.y);
   }
   // update the Grain's position.  First, store where it is now into
   // oldCenter.  Take the velocity from the last update and scale it
   // to simulate drag due to the effects of atmosphere.  Add in
   // gravity.  Then add in the total contribution of all wind gusts.
   // move the Grain to the new position.  If it's fallen off the bottom
   // of the screen, recycle it by loading it with new random values.
   void move() {
      // save the current center
      oldCenter.x = center.x;
      oldCenter.y = center.y;
      // reduce the old velocity to account for atmospheric drag
      PVector newVelocity = new PVector(oldVelocity.x, oldVelocity.y);
      newVelocity.mult(.3);
      // start the new effect (in totalWind) with some gravity
      PVector totalWind = new PVector(0, random(1.5, 2.0));
      // add in the effects of each wind gust
      for (int i=0; i<Gusts.length; i++) {
         PVector w = Gusts[i].windAt(center);
         totalWind.add(w);
      }
      // add the new effect (in totalWind) to the old velocity,
      // giving a new velocity.  Add that to the center.
      newVelocity.add(totalWind);
      center.add(newVelocity);
      oldVelocity.x = newVelocity.x;
      oldVelocity.y = newVelocity.y;
      // if the Grain has fallen below the screen, recycle it.
      if (center.y > height) {
         fillGrain();
      }
   }
}
//----------------------------- WIND CLASS
class Wind {
   PVector center;    // center of the wind
   PVector velocity;  // speed of the wind itself across the screen
   PVector wind;      // direction of the wind
   float radius;      // radius of wind effect
   float strength;    // strength of the wind
   // create a new object with random values
   Wind() {
      center = new PVector(random(50, width-50), random(50, height-50));
      velocity= new PVector(random(-1, 1), random(-1, 1));
      wind= new PVector(random(-3, 3), random(-1, 1));
      radius = random(40, 220);
      strength = random(-5, 5);
   }
   // find the effect of this wind at this point
   PVector windAt(PVector p) {
      PVector windSum = new PVector(0, 0);  // by default, there's no effect
      float pointToWind = dist(p.x, p.y, center.x, center.y);
      if (pointToWind < radius) {
         // we're inside radius.  Find the strength from the distance
         // of this point to the center.
         float wstrength = map(pointToWind, 0, radius, strength, 0);
         // compute the line perpendicular to the line from
         // the center to this point; that's where the wind is pointing.
         PVector dPoint = new PVector(p.x-center.x, p.y-center.y);
         PVector windDir = new PVector(dPoint.y, -dPoint.x);
         // scale the wind based on its strength at this point
         windDir.normalize();
         windDir.mult(wstrength);
         windSum.add(windDir);
      }
      return(windSum);
   }
   // move the wind around on the screen like a bouncing ball
   void update() {
      center.add(velocity);
      if ((center.x < 0) || (center.x >= width)) {
         velocity.x = -velocity.x;
         center.x += 2*velocity.x;
      }
      if ((center.y < 0) || (center.y >= height)) {
         velocity.y = -velocity.y;
         center.y += 2*velocity.y;
      }
   }
}
